package com.smith9.geiger.client.beans

import org.joda.time.DateTime
import simplygroovy.util.DateUtil
import simplygroovy.bean.Period
import org.apache.commons.beanutils.BeanUtils

/**
 * The AccumulatedCountLists is a group of lists, one for each period type.
 * Counts added to this list are levelled, for example adding a count for YEAR with periodStart 18/5/2012 will be levelled to the start of the year
 * i.e. 1/1/2012.
 * <br/>
 * <b>Counts will usually be added using method mergeAccumulate.</b> This means that adding a count of 10 for a MINUTE period will also add a
 * count of 10 to the counts for HOUR, DAY, MONTH and YEAR.
 * <br/>
 * An example use is to keeps track of counts for many different time periods, you might start of processing a feed of counts that have timings down
 * to the millisecond. You then want to be able to get how many counts occurred in any one MINUTE, HOUR, DAY, MONTH or YEAR.
 * This is achieved via accumulation.
 * <br/>
 If your feed was (not all count fields shown)<br/>
 counts.mergeAccumulate(new Count(subCategory: "email", period: Period.MINUTE, periodStart: new DateTime(2012, 5, 19, 21, 21, 47, 111), count: 1))<br/>
 counts.mergeAccumulate(new Count(subCategory: "email", period: Period.MINUTE, periodStart: new DateTime(2012, 5, 19, 21, 21, 48, 111), count: 1))<br/>
 counts.mergeAccumulate(new Count(subCategory: "email", period: Period.MINUTE, periodStart: new DateTime(2012, 5, 19, 21, 21, 49, 111), count: 1))<br/>
 counts.mergeAccumulate(new Count(subCategory: "email", period: Period.MINUTE, periodStart: new DateTime(2011, 4, 18, 14, 16, 49, 111), count: 1))<br/>
 counts.mergeAccumulate(new Count(subCategory: "email", period: Period.MINUTE, periodStart: new DateTime(2011, 5, 18, 14, 16, 49, 111), count: 1))<br/>

<table>
  <tr>
    <th>SubCategory</th>
    <th>Count</th>
    <th>Period</th>
    <th>PeriodStart</th>
  </tr>
  <tr>
    <td>email</td>
    <td>3</td>
    <td>MINUTE</td>
    <td>19/05/2012 21:21</td>
  </tr>
  <tr>
    <td>email</td>
    <td>1</td>
    <td>MINUTE</td>
    <td>18/04/2012 14:16</td>
  </tr>
  <tr>
    <td>email</td>
    <td>3</td>
    <td>HOUR</td>
    <td>19/05/2012 21:00</td>
  </tr>
  <tr>
    <td>email</td>
    <td>1</td>
    <td>HOUR</td>
    <td>18/04/2012 14:00</td>
  </tr>
  <tr>
    <td>email</td>
    <td>3</td>
    <td>DAY</td>
    <td>19/05/2012 00:00</td>
  </tr>
  <tr>
    <td>email</td>
    <td>1</td>
    <td>DAY</td>
    <td>18/04/2012 00:00</td>
  </tr>
  <tr>
    <td>email</td>
    <td>3</td>
    <td>MONTH</td>
    <td>01/05/2012 00:00</td>
  </tr>
  <tr>
    <td>email</td>
    <td>1</td>
    <td>MONTH</td>
    <td>01/04/2012 00:00</td>
  </tr>
  <tr>
    <td>email</td>
    <td>4</td>
    <td>YEAR</td>
    <td>01/01/2012 00:00</td>
  </tr>
</table>
 * To load the counts into Geiger, you can use GeigerClient, BatchingGeigerClient and write your own loader or use
 * GeigerLoader where you just supply your own csv or xml data
 * Created: 18/05/12
 * @author david
 */
class AccumulatedCountLists {

    private Map<Period, CountList> countLists = [:].withDefault {new CountList()}

    public void addSizeTrigger(int threshold, Period period, Closure sizeReachedAction) {
        getCountList(period).addSizeTrigger(threshold, sizeReachedAction)
    }

    public void addSizeTrigger(int threshold, Closure sizeReachedAction) {
        Period.values().each { Period period ->
            getCountList(period).addSizeTrigger(threshold, sizeReachedAction)
        }
    }

    /**
     * collects all counts from all of the sub lists into a single list
     * @param periods
     * @return
     */
    public CountList getAll(Period ... periods = Period.values()) {
        CountList allCounts = new CountList()
        periods.each { Period period ->
            allCounts.addAll(countLists[period].getCountValues())
        }
        return allCounts
    }
    /**
     * get the sublist for the given period
     * @param period
     * @return
     */
    public CountList getCountList(Period period) {
        return countLists[period]
    }
    /**
     * determine if all of the lists for the given periods are empty
     * @param periods
     * @return
     */
    public boolean isEmpty(Period ... periods = Period.values()) {
        boolean isEmpty = true
        Iterator<CountList> iter = countLists.values().iterator()
        while(iter.hasNext()) {
            if(!iter.next().isEmpty()) {
                isEmpty = false
                break;
            }
        }
        return isEmpty
    }
    /**
     * clears the counts from all of the lists
     */
    public void clear() {
        countLists.values().each { it.clear() }
    }
    /**
     * total of all counts in all of the lists
     * @return
     */
    public int size() {
        int size = 0
        countLists.values().each { size += it.size() }
        return size
    }
    /**
     * removes a count based on the fields used in Count.key()
     * periodStart will be levelled.
     * @param count
     * @return
     */
    public Count remove(Count count) {
        CountList counts = countLists[count.period]
        return counts.remove(count)
    }
    /**
     * Use this to add counts without merging or accumulating. You would normally use mergeAccumulate to add counts.
     * For example if the list contain a count of 10 for period 9:03 and you add a count of 5 for period 9:03
     * you end up with two counts for period 9:03
     * periodStart will be levelled.
     * @param count
     */
    public void add(Count count) {
        CountList counts = countLists[count.period]
        counts.add(count)
    }
    /**
     * Will merge totals with counts that match the same time period.
     * For example if a list contains a count of 10 for 9:03 and you add a count for 9:03 of 5
     * then you end up with a single count of 15 for 9:03.
     * Counts are also accumulated. For example, when you add a count of 10 for a minute period
     * a count of 10 is also added for periods hour, day, month and year
     * periodStart will be levelled.
     * @param count
     */
    public void mergeAccumulate(Count count) {
        switch(count.period) {
            case Period.MINUTE:
                merge(count.createCopyLevelDate(Period.MINUTE))
            case Period.HOUR:
                merge(count.createCopyLevelDate(Period.HOUR))
            case Period.DAY:
                merge(count.createCopyLevelDate(Period.DAY))
            case Period.MONTH:
                merge(count.createCopyLevelDate(Period.MONTH))
            case Period.YEAR:
                merge(count.createCopyLevelDate(Period.YEAR))
        }
    }

    /**
     * Will merge totals with counts that match the same time period.
     * For example if a list contains a count of 10 for 9:03 and you add a count for 9:03 of 5
     * then you end up with a single count of 15 for 9:03.
     * periodStart will be levelled.
     * @param count
     */
    public void merge(Count count) {
        CountList counts = countLists[count.period]
        counts.merge(count)
    }

    /**
     * finds a count based on the fields in count (environment, productName, subProduct, period, category, subCategory, failure, periodStart)
     * This is the values returned by Count.key()
     * If the periodStart is not exactly the start of the period, it will levelled to achieve the find.
     * For example, if the list contains a count for YEAR 2012 is searched for using a date of 18/5/2012, then the date is levelled to 1/1/2012
     * so that the count can be found
     * @param count used to search with, fields returned in Count.key() are used in the search.
     * @return
     */
    public Count find(Count count) {
        return countLists[count.period].find(count)
    }
}
